import * as path from 'path';
import express from 'express';
import cookieParser from 'cookie-parser';
import { fs, createDebugger, compatRequire } from '@modern-js/utils';
import finalhandler from 'finalhandler';
import { run } from "./context";
import registerRoutes from "./registerRoutes";
const debug = createDebugger('express');

const findAppModule = async apiDir => {
  const exts = ['.ts', '.js'];
  const paths = exts.map(ext => path.resolve(apiDir, `app${ext}`));

  for (const filename of paths) {
    if (await fs.pathExists(filename)) {
      // 每次获取 app.ts 的时候，避免使用缓存的 app.ts
      delete require.cache[filename];
      return [compatRequire(filename), require(filename)];
    }
  }

  return [];
};

const initMiddlewares = (middleware, app) => {
  middleware.forEach(middlewareItem => {
    const middlewareFunc = typeof middlewareItem === 'string' ? compatRequire(middlewareItem) : middlewareItem;
    app.use(middlewareFunc);
  });
};

const useRun = app => {
  app.use((req, res, next) => {
    run({
      req,
      res
    }, next);
  });
};

const initApp = app => {
  app.use(cookieParser());
  app.use(express.text());
  app.use(express.urlencoded({
    extended: true
  }));
  app.use(express.json());
  return app;
};

export default (() => ({
  name: '@modern-js/plugin-express',
  pre: ['@modern-js/plugin-bff'],
  setup: api => ({
    async prepareApiServer({
      pwd,
      config
    }) {
      let app;
      const apiDir = path.join(pwd, './api');
      const appContext = api.useAppContext();
      const apiHandlerInfos = appContext.apiHandlerInfos;
      const mode = appContext.apiMode;

      if (mode === 'framework') {
        const appModule = await findAppModule(apiDir);
        app = appModule[0];
        const hooks = appModule[1];

        if (!app || !app.use) {
          // console.warn('There is not api/app.ts.');
          app = express();
        }

        initApp(app);

        if (config) {
          const {
            middleware
          } = config;
          initMiddlewares(middleware, app);
        }

        useRun(app);
        registerRoutes(app, apiHandlerInfos);

        if (hooks) {
          const {
            afterLambdaRegisted
          } = hooks;

          if (afterLambdaRegisted) {
            afterLambdaRegisted(app);
          }
        }
      } else if (mode === 'function') {
        app = express();
        initApp(app);

        if (config) {
          const {
            middleware
          } = config;
          initMiddlewares(middleware, app);
        }

        useRun(app);
        registerRoutes(app, apiHandlerInfos);
      } else {
        throw new Error(`mode must be function or framework`);
      }

      return (req, res) => new Promise((resolve, reject) => {
        const handler = err => {
          if (err) {
            return reject(err);
          } // finalhanlder will trigger 'finish' event


          return finalhandler(req, res, {})(null); // return resolve();
        };

        res.on('finish', err => {
          if (err) {
            return reject(err);
          }

          return resolve();
        });
        return app(req, res, handler);
      });
    },

    prepareWebServer({
      config
    }) {
      const app = express();
      initApp(app);

      if (config) {
        const {
          middleware
        } = config;
        debug('web middleware', middleware);
        initMiddlewares(middleware, app);
      }

      return (req, res) => new Promise((resolve, reject) => {
        const handler = err => {
          if (err) {
            return reject(err);
          }

          if (res.headersSent && res.statusCode !== 200) {
            finalhandler(req, res, {})(null);
          }

          return resolve();
        }; // when user call res.send


        res.on('finish', err => {
          if (err) {
            return reject(err);
          }

          return resolve();
        });
        return app(req, res, handler);
      });
    }

  })
}));